==> haystack-0.6.1.tar.gz: lib/src/io/HaystackTokenizer.dart <==
///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.tokenizer;

import 'package:haystack/src/HDate.dart';
import 'package:haystack/src/HDateTime.dart';
import 'package:haystack/src/HNum.dart';
import 'package:haystack/src/HRef.dart';
import 'package:haystack/src/HStr.dart';
import 'package:haystack/src/HTime.dart';
import 'package:haystack/src/HUri.dart';
import 'package:haystack/src/ParseError.dart';
import 'package:haystack/src/string_utils.dart';
import 'package:unicode/unicode.dart' as unicode;

import 'HaystackToken.dart';

/// Stream based tokenizer for Haystack formats such as Zinc and Filters.
class HaystackTokenizer {
//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////
  static final int eof = -1;
  HaystackToken tok; //  _current token type
  Object val; //token literal or identifier
  int line = 1; //  _current line number

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  HaystackTokenizer(Iterable<int> inStream)
      : this._inStream = inStream.iterator {
    this.tok = HaystackToken.eof;
    _consume();
    _consume();
  }

  bool close() {
//    try {
//      //_inStream.close();
    return true;
//    } catch (IOException e) {
//      return false;
//    }
  }

//////////////////////////////////////////////////////////////////////////
// Tokenizing
//////////////////////////////////////////////////////////////////////////

  HaystackToken next() {
    // reset
    val = null;

    // skip non-meaningful whitespace and comments
    while (true) {
      // treat space, tab, non-breaking space as whitespace
      if (_cur == SPACE_CHAR || _cur == TAB_CHAR || _cur == 0xa0) {
        _consume();
        continue;
      }

      // comments
      if (_cur == SLASH_CHAR) {
        if (_peek == SLASH_CHAR) {
          _skipCommentsSL();
          continue;
        }
        if (_peek == ASTERISK_CHAR) {
          _skipCommentsML();
          continue;
        }
      }

      break;
    }

    // newlines
    if (_cur == NL_CHAR || _cur == CR_CHAR) {
      if (_cur == CR_CHAR && _peek == NL_CHAR) _consume('\r');
      _consume();
      ++line;
      return tok = HaystackToken.nl;
    }

    // handle various starting chars
    if (_isIdStart(_cur)) return tok = _id();
    if (_cur == DOUBLE_QUOTE_CHAR) return tok = _str();
    if (_cur == DOG_CHAR) return tok = _ref();
    if (_isDigit(_cur)) return tok = _num();
    if (_cur == BACKQUOTE_CHAR) return tok = _uri();
    if (_cur == MINUS_CHAR && _isDigit(_peek)) return tok = _num();

    return tok = _symbol();
  }

//////////////////////////////////////////////////////////////////////////
// Token Productions
//////////////////////////////////////////////////////////////////////////

  HaystackToken _id() {
    StringBuffer s = new StringBuffer();

    while (_isIdPart(_cur)) {
      s.writeCharCode(_cur);
      _consume();
    }

    this.val = s.toString();
    return HaystackToken.id;
  }

  static bool _isIdStart(int _cur) {
    if (a_CHAR <= _cur && _cur <= z_CHAR) return true;
    if (A_CHAR <= _cur && _cur <= Z_CHAR) return true;
    return false;
  }

  static bool _isIdPart(int _cur) {
    if (_isIdStart(_cur)) return true;
    if (_isDigit(_cur)) return true;
    if (_cur == UNDERSCORE_CHAR) return true;
    return false;
  }

  static bool _isDigit(int c) => (c >= ZERO_CHAR && c <= NINE_CHAR);
  static bool _isLetter(int c) => (c != eof && isLetter(c));

  HaystackToken _num() {
    // hex number (no unit allowed)
    bool isHex = (_cur == ZERO_CHAR && _peek == x_CHAR);
    if (isHex) {
      _consume('0');
      _consume('x');
      StringBuffer s = new StringBuffer();

      while (true) {
        if (_isHex(_cur)) {
          s.writeCharCode(_cur);
          _consume();
          continue;
        }
        if (_cur == UNDERSCORE_CHAR) {
          _consume();
          continue;
        }
        break;
      }

      this.val = new HNum(int.parse(s.toString(), radix: 16));
      return HaystackToken.num;
    }

    //  _consume all things that might be part of this number token
    StringBuffer s = new StringBuffer()..writeCharCode(_cur);
    _consume();
    int colons = 0;
    int dashes = 0;
    int unitIndex = 0;
    bool exp = false;

    while (true) {
      if (!_isDigit(_cur)) {
        if (exp && (_cur == PLUS_CHAR || _cur == MINUS_CHAR)) {
        } else if (_cur == MINUS_CHAR) {
          ++dashes;
        } else if (_cur == COLON_CHAR && _isDigit(_peek)) {
          ++colons;
        } else if ((exp || colons >= 1) && _cur == PLUS_CHAR) {
        } else if (_cur == PERIOD_CHAR) {
          if (!_isDigit(_peek)) break;
        } else if ((_cur == e_CHAR || _cur == E_CHAR) &&
            (_peek == MINUS_CHAR || _peek == PLUS_CHAR || _isDigit(_peek))) {
          exp = true;
        } else if (_isLetter(_cur) ||
            _cur == PERCENT_CHAR ||
            _cur == DOLLAR_CHAR ||
            _cur == SLASH_CHAR ||
            _cur > 128) {
          if (unitIndex == 0) unitIndex = s.length;
        } else if (_cur == UNDERSCORE_CHAR) {
          if (unitIndex == 0 && _isDigit(_peek)) {
            _consume();
            continue;
          } else {
            if (unitIndex == 0) unitIndex = s.length;
          }
        } else {
          break;
        }
      }

      s.writeCharCode(_cur);
      _consume();
    }

    if (dashes == 2 && colons == 0) return _date(s.toString());
    if (dashes == 0 && colons >= 1) return _time(s, colons == 1);
    if (dashes >= 2) return _dateTime(s);
    return _number(s.toString(), unitIndex);
  }

  static bool _isHex(int c) {
    if (c == eof) return false;
    c = new String.fromCharCode(c).toLowerCase().codeUnitAt(0);
    if (a_CHAR <= c && c <= f_CHAR) return true;
    if (_isDigit(c)) return true;
    return false;
  }

  HaystackToken _date(String s) {
    try {
      this.val = new HDate.fromString(s);
      return HaystackToken.date;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  /// we don't require hour to be two digits and we don't require seconds
  HaystackToken _time(StringBuffer s, bool addSeconds) {
    try {
      var str = s.toString();

      if (str[1] == ":") str = "0$str";
      if (addSeconds) str = "$str:00";
      this.val = new HTime.fromString(str);
      return HaystackToken.time;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  HaystackToken _dateTime(StringBuffer s) {
    // xxx timezone
    if (_cur != SPACE_CHAR || !unicode.isUppercaseLetter(_peek)) {
      if (s.toString()[s.length - 1] == 'Z')
        s.write(" UTC");
      else
        throw _err("Expecting timezone");
    } else {
      _consume();
      s.write(' ');
      while (_isIdPart(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      }

      // handle GMT+xx or GMT-xx
      if ((_cur == PLUS_CHAR || _cur == MINUS_CHAR) &&
          s.toString().endsWith("GMT")) {
        s.writeCharCode(_cur);
        _consume();
        while (_isDigit(_cur)) {
          s.writeCharCode(_cur);
          _consume();
        }
      }
    }

    try {
      this.val = new HDateTime.fromString(s.toString());
      return HaystackToken.dateTime;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  HaystackToken _number(String s, int unitIndex) {
    try {
      if (unitIndex == 0) {
        this.val = new HNum(num.parse(s));
      } else {
        String doubleStr = s.substring(0, unitIndex);
        String unitStr = s.substring(unitIndex);
        this.val = new HNum(num.parse(doubleStr), unitStr);
      }
    } catch (e) {
      throw _err("Invalid Number literal: $s");
    }

    return HaystackToken.num;
  }

  HaystackToken _str() {
    _consume('"');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (_cur == eof) throw _err("Unexpected end of str");
      if (_cur == DOUBLE_QUOTE_CHAR) {
        _consume('"');
        break;
      }
      if (_cur == BACKSLASH_CHAR) {
        s.writeCharCode(_escape());
        continue;
      }
      s.writeCharCode(_cur);
      _consume();
    }

    this.val = new HStr(s.toString());
    return HaystackToken.str;
  }

  HaystackToken _ref() {
    _consume('@');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (HRef.isIdChar(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      } else {
        break;
      }
    }

    this.val = new HRef(s.toString(), null);
    return HaystackToken.ref;
  }

  HaystackToken _uri() {
    _consume('`');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (_cur == BACKQUOTE_CHAR) {
        _consume('`');
        break;
      }
      if (_cur == eof || _cur == NL_CHAR) throw _err("Unexpected end of uri");
      if (_cur == BACKSLASH_CHAR) {
        /// TODO or convert all chars to int ASC codes in below cases:
        switch (_peek) {
          case COLON_CHAR:
          case SLASH_CHAR:
          case QUESTION_MARK_CHAR:
          case NUMBER_SIGN_CHAR:
          case OPEN_BRACKET_CHAR:
          case CLOSE_BRACKET_CHAR:
          case DOG_CHAR:
          case BACKSLASH_CHAR:
          case AMPERSAND_CHAR:
          case EQUALS_SIGN_CHAR:
          case SEMICOLON_CHAR:
            s.writeCharCode(_cur);
            s.writeCharCode(_peek);
            _consume();
            _consume();
            break;
          default:
            s.writeCharCode(_escape());
        }
      } else {
        s.writeCharCode(_cur);
        _consume();
      }
    }

    this.val = new HUri(s.toString());
    return HaystackToken.uri;
  }

  int _escape() {
    _consume('\\');
    switch (_cur) {
      case b_CHAR:
        _consume();
        return BS_CHAR; // \b
      case f_CHAR:
        _consume();
        return FF_CHAR; // \f
      case n_CHAR:
        _consume();
        return NL_CHAR;
      case r_CHAR:
        _consume();
        return CR_CHAR;
      case t_CHAR:
        _consume();
        return TAB_CHAR;
      case DOUBLE_QUOTE_CHAR:
        _consume();
        return DOUBLE_QUOTE_CHAR;
      case DOLLAR_CHAR:
        _consume();
        return DOLLAR_CHAR;
      case QUOTE_CHAR:
        _consume();
        return '\''.codeUnitAt(0);
      case BACKQUOTE_CHAR:
        _consume();
        return BACKQUOTE_CHAR;
      case BACKSLASH_CHAR:
        _consume();
        return BACKSLASH_CHAR;
    }

    // check for uxxxx
    StringBuffer esc = new StringBuffer();
    if (_cur == u_CHAR) {
      _consume('u');
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();

      try {
        return int.parse(esc.toString(), radix: 16);
      } on FormatException {
        throw new ParseError("Invalid unicode escape: $esc");
      }
    }

    throw _err("Invalid escape sequence: " + new String.fromCharCode(_cur));
  }

  HaystackToken _symbol() {
    int c = _cur;
    _consume();
    switch (c) {
      case COMMA_CHAR:
        return HaystackToken.comma;
      case COLON_CHAR:
        return HaystackToken.colon;
      case SEMICOLON_CHAR: // ;
        return HaystackToken.semicolon;
      case OPEN_BRACKET_CHAR: // [
        return HaystackToken.lbracket;
      case CLOSE_BRACKET_CHAR: //  ]
        return HaystackToken.rbracket;
      case OPEN_BRACE_CHAR: // {
        return HaystackToken.lbrace;
      case CLOSE_BRACE_CHAR: // }
        return HaystackToken.rbrace;
      case OPEN_PARENTHESE_CHAR: // (
        return HaystackToken.lparen;
      case CLOSE_PARENTHESE_CHAR: // )
        return HaystackToken.rparen;
      case LESS_CHAR: // <
        if (_cur == LESS_CHAR) {
          _consume('<');
          return HaystackToken.lt2;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.ltEq;
        }
        return HaystackToken.lt;
      case GREATER_CHAR: // >
        if (_cur == GREATER_CHAR) {
          _consume('>');
          return HaystackToken.gt2;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.gtEq;
        }
        return HaystackToken.gt;
      case MINUS_CHAR: // -
        if (_cur == GREATER_CHAR) {
          _consume('>');
          return HaystackToken.arrow;
        }
        return HaystackToken.minus;
      case EQUALS_SIGN_CHAR: // =
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.eq;
        }
        return HaystackToken.assign;
      case EXCLAMATION_CHAR: // !
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.notEq;
        }
        return HaystackToken.bang;
      case SLASH_CHAR: // /
        return HaystackToken.slash;
    }

    if (c == eof) return HaystackToken.eof;
    throw _err("Unexpected symbol: '" +
        new String.fromCharCode(c) +
        "' (0x" +
        c.toRadixString(16) +
        ")");
  }

//////////////////////////////////////////////////////////////////////////
// Comments
//////////////////////////////////////////////////////////////////////////

  void _skipCommentsSL() {
    _consume('/');
    _consume('/');
    while (true) {
      if (_cur == NL_CHAR || _cur == eof) break;
      _consume();
    }
  }

  void _skipCommentsML() {
    _consume('/');
    _consume('*');
    int depth = 1;
    while (true) {
      if (_cur == ASTERISK_CHAR && _peek == SLASH_CHAR) {
        _consume('*');
        _consume('/');
        depth--;
        if (depth <= 0) break;
      }
      if (_cur == SLASH_CHAR && _peek == ASTERISK_CHAR) {
        _consume('/');
        _consume('*');
        depth++;
        continue;
      }
      if (_cur == NL_CHAR) ++line;
      if (_cur == eof) throw _err("Multi-line comment not closed");
      _consume();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Error Handling
//////////////////////////////////////////////////////////////////////////

  ParseError _err(String msg) => new ParseError("$msg [line $line]");

//////////////////////////////////////////////////////////////////////////
// Char
//////////////////////////////////////////////////////////////////////////

  void _consume([String expected]) {
    if (expected != null && _cur != expected.codeUnitAt(0))
      throw _err("Expected $expected");

    try {
      _cur = _peek;
      _inStream.moveNext();
      _peek = _inStream.current ?? eof;
    } catch (e) {
      _cur = eof;
      _peek = eof;
    }
  }

  final Iterator<int> _inStream; // underlying stream
  int _cur = eof; //  _current char
  int _peek = eof; // next char
}

==> haystack-1.0.0-dev.0.3.tar.gz: lib/src/io/HaystackTokenizer.dart <==
///
/// Copyright (c) 2019, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.tokenizer;

import 'package:haystack/src/HDate.dart';
import 'package:haystack/src/HDateTime.dart';
import 'package:haystack/src/HNum.dart';
import 'package:haystack/src/HRef.dart';
import 'package:haystack/src/HStr.dart';
import 'package:haystack/src/HTime.dart';
import 'package:haystack/src/HUri.dart';
import 'package:haystack/src/ParseError.dart';
import 'package:haystack/src/string_utils.dart';
import 'package:unicode/unicode.dart' as unicode;

import 'HaystackToken.dart';

/// Stream based tokenizer for Haystack formats such as Zinc and Filters.
class HaystackTokenizer {
//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////
  static final int eof = -1;
  HaystackToken tok; //  _current token type
  Object val; //token literal or identifier
  int line = 1; //  _current line number

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  HaystackTokenizer(Iterable<int> inStream)
      : this._inStream = inStream.iterator {
    this.tok = HaystackToken.eof;
    _consume();
    _consume();
  }

  bool close() {
//    try {
//      //_inStream.close();
    return true;
//    } catch (IOException e) {
//      return false;
//    }
  }

//////////////////////////////////////////////////////////////////////////
// Tokenizing
//////////////////////////////////////////////////////////////////////////

  HaystackToken next() {
    // reset
    val = null;

    // skip non-meaningful whitespace and comments
    while (true) {
      // treat space, tab, non-breaking space as whitespace
      if (_cur == SPACE_CHAR || _cur == TAB_CHAR || _cur == 0xa0) {
        _consume();
        continue;
      }

      // comments
      if (_cur == SLASH_CHAR) {
        if (_peek == SLASH_CHAR) {
          _skipCommentsSL();
          continue;
        }
        if (_peek == ASTERISK_CHAR) {
          _skipCommentsML();
          continue;
        }
      }

      break;
    }

    // newlines
    if (_cur == NL_CHAR || _cur == CR_CHAR) {
      if (_cur == CR_CHAR && _peek == NL_CHAR) {
        _consume('\r');
      }
      _consume();
      ++line;
      return tok = HaystackToken.nl;
    }

    // handle various starting chars
    if (_isIdStart(_cur)) return tok = _id();
    if (_cur == DOUBLE_QUOTE_CHAR) return tok = _str();
    if (_cur == DOG_CHAR) return tok = _ref();
    if (_isDigit(_cur)) return tok = _num();
    if (_cur == BACKQUOTE_CHAR) return tok = _uri();
    if (_cur == MINUS_CHAR && _isDigit(_peek)) return tok = _num();

    return tok = _symbol();
  }

//////////////////////////////////////////////////////////////////////////
// Token Productions
//////////////////////////////////////////////////////////////////////////

  HaystackToken _id() {
    StringBuffer s = new StringBuffer();

    while (_isIdPart(_cur)) {
      s.writeCharCode(_cur);
      _consume();
    }

    this.val = s.toString();
    return HaystackToken.id;
  }

  static bool _isIdStart(int _cur) {
    if (a_CHAR <= _cur && _cur <= z_CHAR) return true;
    if (A_CHAR <= _cur && _cur <= Z_CHAR) return true;
    return false;
  }

  static bool _isIdPart(int _cur) {
    if (_isIdStart(_cur)) return true;
    if (_isDigit(_cur)) return true;
    if (_cur == UNDERSCORE_CHAR) return true;
    return false;
  }

  static bool _isDigit(int c) => (c >= ZERO_CHAR && c <= NINE_CHAR);
  static bool _isLetter(int c) => (c != eof && isLetter(c));

  HaystackToken _num() {
    // hex number (no unit allowed)
    bool isHex = (_cur == ZERO_CHAR && _peek == x_CHAR);
    if (isHex) {
      _consume('0');
      _consume('x');
      StringBuffer s = new StringBuffer();

      while (true) {
        if (_isHex(_cur)) {
          s.writeCharCode(_cur);
          _consume();
          continue;
        }
        if (_cur == UNDERSCORE_CHAR) {
          _consume();
          continue;
        }
        break;
      }

      this.val = new HNum(int.parse(s.toString(), radix: 16));
      return HaystackToken.num;
    }

    //  _consume all things that might be part of this number token
    StringBuffer s = new StringBuffer()..writeCharCode(_cur);
    _consume();
    int colons = 0;
    int dashes = 0;
    int unitIndex = 0;
    bool exp = false;

    while (true) {
      if (!_isDigit(_cur)) {
        if (exp && (_cur == PLUS_CHAR || _cur == MINUS_CHAR)) {
        } else if (_cur == MINUS_CHAR) {
          ++dashes;
        } else if (_cur == COLON_CHAR && _isDigit(_peek)) {
          ++colons;
        } else if ((exp || colons >= 1) && _cur == PLUS_CHAR) {
        } else if (_cur == PERIOD_CHAR) {
          if (!_isDigit(_peek)) break;
        } else if ((_cur == e_CHAR || _cur == E_CHAR) &&
            (_peek == MINUS_CHAR || _peek == PLUS_CHAR || _isDigit(_peek))) {
          exp = true;
        } else if (_isLetter(_cur) ||
            _cur == PERCENT_CHAR ||
            _cur == DOLLAR_CHAR ||
            _cur == SLASH_CHAR ||
            _cur > 128) {
          if (unitIndex == 0) {
            unitIndex = s.length;
          }
        } else if (_cur == UNDERSCORE_CHAR) {
          if (unitIndex == 0 && _isDigit(_peek)) {
            _consume();
            continue;
          } else {
            if (unitIndex == 0) {
              unitIndex = s.length;
            }
          }
        } else {
          break;
        }
      }

      s.writeCharCode(_cur);
      _consume();
    }

    if (dashes == 2 && colons == 0) return _date(s.toString());
    if (dashes == 0 && colons >= 1) return _time(s, colons == 1);
    if (dashes >= 2) return _dateTime(s);
    return _number(s.toString(), unitIndex);
  }

  static bool _isHex(int c) {
    if (c == eof) return false;
    c = new String.fromCharCode(c).toLowerCase().codeUnitAt(0);
    if (a_CHAR <= c && c <= f_CHAR) return true;
    if (_isDigit(c)) return true;
    return false;
  }

  HaystackToken _date(String s) {
    try {
      this.val = new HDate.fromString(s);
      return HaystackToken.date;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  /// we don't require hour to be two digits and we don't require seconds
  HaystackToken _time(StringBuffer s, bool addSeconds) {
    try {
      var str = s.toString();

      if (str[1] == ":") {
        str = "0$str";
      }
      if (addSeconds) {
        str = "$str:00";
      }
      this.val = new HTime.fromString(str);
      return HaystackToken.time;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  HaystackToken _dateTime(StringBuffer s) {
    // xxx timezone
    if (_cur != SPACE_CHAR || !unicode.isUppercaseLetter(_peek)) {
      if (s.toString()[s.length - 1] == 'Z') {
        s.write(" UTC");
      } else {
        throw _err("Expecting timezone");
      }
    } else {
      _consume();
      s.write(' ');
      while (_isIdPart(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      }

      // handle GMT+xx or GMT-xx
      if ((_cur == PLUS_CHAR || _cur == MINUS_CHAR) &&
          s.toString().endsWith("GMT")) {
        s.writeCharCode(_cur);
        _consume();
        while (_isDigit(_cur)) {
          s.writeCharCode(_cur);
          _consume();
        }
      }
    }

    try {
      this.val = new HDateTime.fromString(s.toString());
      return HaystackToken.dateTime;
    } catch (e) {
      throw _err(e.toString());
    }
  }

  HaystackToken _number(String s, int unitIndex) {
    try {
      if (unitIndex == 0) {
        this.val = new HNum(num.parse(s));
      } else {
        String doubleStr = s.substring(0, unitIndex);
        String unitStr = s.substring(unitIndex);
        this.val = new HNum(num.parse(doubleStr), unitStr);
      }
    } catch (e) {
      throw _err("Invalid Number literal: $s");
    }

    return HaystackToken.num;
  }

  HaystackToken _str() {
    _consume('"');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (_cur == eof) throw _err("Unexpected end of str");
      if (_cur == DOUBLE_QUOTE_CHAR) {
        _consume('"');
        break;
      }
      if (_cur == BACKSLASH_CHAR) {
        s.writeCharCode(_escape());
        continue;
      }
      s.writeCharCode(_cur);
      _consume();
    }

    this.val = new HStr(s.toString());
    return HaystackToken.str;
  }

  HaystackToken _ref() {
    _consume('@');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (HRef.isIdChar(_cur)) {
        s.writeCharCode(_cur);
        _consume();
      } else {
        break;
      }
    }

    this.val = new HRef(s.toString(), null);
    return HaystackToken.ref;
  }

  HaystackToken _uri() {
    _consume('`');
    StringBuffer s = new StringBuffer();

    while (true) {
      if (_cur == BACKQUOTE_CHAR) {
        _consume('`');
        break;
      }
      if (_cur == eof || _cur == NL_CHAR) throw _err("Unexpected end of uri");
      if (_cur == BACKSLASH_CHAR) {
        /// TODO or convert all chars to int ASC codes in below cases:
        switch (_peek) {
          case COLON_CHAR:
          case SLASH_CHAR:
          case QUESTION_MARK_CHAR:
          case NUMBER_SIGN_CHAR:
          case OPEN_BRACKET_CHAR:
          case CLOSE_BRACKET_CHAR:
          case DOG_CHAR:
          case BACKSLASH_CHAR:
          case AMPERSAND_CHAR:
          case EQUALS_SIGN_CHAR:
          case SEMICOLON_CHAR:
            s.writeCharCode(_cur);
            s.writeCharCode(_peek);
            _consume();
            _consume();
            break;
          default:
            s.writeCharCode(_escape());
        }
      } else {
        s.writeCharCode(_cur);
        _consume();
      }
    }

    this.val = new HUri(s.toString());
    return HaystackToken.uri;
  }

  int _escape() {
    _consume('\\');
    switch (_cur) {
      case b_CHAR:
        _consume();
        return BS_CHAR; // \b
      case f_CHAR:
        _consume();
        return FF_CHAR; // \f
      case n_CHAR:
        _consume();
        return NL_CHAR;
      case r_CHAR:
        _consume();
        return CR_CHAR;
      case t_CHAR:
        _consume();
        return TAB_CHAR;
      case DOUBLE_QUOTE_CHAR:
        _consume();
        return DOUBLE_QUOTE_CHAR;
      case DOLLAR_CHAR:
        _consume();
        return DOLLAR_CHAR;
      case QUOTE_CHAR:
        _consume();
        return '\''.codeUnitAt(0);
      case BACKQUOTE_CHAR:
        _consume();
        return BACKQUOTE_CHAR;
      case BACKSLASH_CHAR:
        _consume();
        return BACKSLASH_CHAR;
    }

    // check for uxxxx
    StringBuffer esc = new StringBuffer();
    if (_cur == u_CHAR) {
      _consume('u');
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();
      esc.writeCharCode(_cur);
      _consume();

      try {
        return int.parse(esc.toString(), radix: 16);
      } on FormatException {
        throw new ParseError("Invalid unicode escape: $esc");
      }
    }

    throw _err("Invalid escape sequence: " + new String.fromCharCode(_cur));
  }

  HaystackToken _symbol() {
    int c = _cur;
    _consume();
    switch (c) {
      case COMMA_CHAR:
        return HaystackToken.comma;
      case COLON_CHAR:
        return HaystackToken.colon;
      case SEMICOLON_CHAR: // ;
        return HaystackToken.semicolon;
      case OPEN_BRACKET_CHAR: // [
        return HaystackToken.lbracket;
      case CLOSE_BRACKET_CHAR: //  ]
        return HaystackToken.rbracket;
      case OPEN_BRACE_CHAR: // {
        return HaystackToken.lbrace;
      case CLOSE_BRACE_CHAR: // }
        return HaystackToken.rbrace;
      case OPEN_PARENTHESE_CHAR: // (
        return HaystackToken.lparen;
      case CLOSE_PARENTHESE_CHAR: // )
        return HaystackToken.rparen;
      case LESS_CHAR: // <
        if (_cur == LESS_CHAR) {
          _consume('<');
          return HaystackToken.lt2;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.ltEq;
        }
        return HaystackToken.lt;
      case GREATER_CHAR: // >
        if (_cur == GREATER_CHAR) {
          _consume('>');
          return HaystackToken.gt2;
        }
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.gtEq;
        }
        return HaystackToken.gt;
      case MINUS_CHAR: // -
        if (_cur == GREATER_CHAR) {
          _consume('>');
          return HaystackToken.arrow;
        }
        return HaystackToken.minus;
      case EQUALS_SIGN_CHAR: // =
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.eq;
        }
        return HaystackToken.assign;
      case EXCLAMATION_CHAR: // !
        if (_cur == EQUALS_SIGN_CHAR) {
          _consume('=');
          return HaystackToken.notEq;
        }
        return HaystackToken.bang;
      case SLASH_CHAR: // /
        return HaystackToken.slash;
    }

    if (c == eof) return HaystackToken.eof;
    throw _err("Unexpected symbol: '" +
        new String.fromCharCode(c) +
        "' (0x" +
        c.toRadixString(16) +
        ")");
  }

//////////////////////////////////////////////////////////////////////////
// Comments
//////////////////////////////////////////////////////////////////////////

  void _skipCommentsSL() {
    _consume('/');
    _consume('/');
    while (true) {
      if (_cur == NL_CHAR || _cur == eof) break;
      _consume();
    }
  }

  void _skipCommentsML() {
    _consume('/');
    _consume('*');
    int depth = 1;
    while (true) {
      if (_cur == ASTERISK_CHAR && _peek == SLASH_CHAR) {
        _consume('*');
        _consume('/');
        depth--;
        if (depth <= 0) break;
      }
      if (_cur == SLASH_CHAR && _peek == ASTERISK_CHAR) {
        _consume('/');
        _consume('*');
        depth++;
        continue;
      }
      if (_cur == NL_CHAR) {
        ++line;
      }
      if (_cur == eof) throw _err("Multi-line comment not closed");
      _consume();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Error Handling
//////////////////////////////////////////////////////////////////////////

  ParseError _err(String msg) => new ParseError("$msg [line $line]");

//////////////////////////////////////////////////////////////////////////
// Char
//////////////////////////////////////////////////////////////////////////

  void _consume([String expected]) {
    if (expected != null && _cur != expected.codeUnitAt(0)) {
      throw _err("Expected $expected");
    }

    try {
      _cur = _peek;
      _inStream.moveNext();
      _peek = _inStream.current ?? eof;
    } catch (e) {
      _cur = eof;
      _peek = eof;
    }
  }

  final Iterator<int> _inStream; // underlying stream
  int _cur = eof; //  _current char
  int _peek = eof; // next char
}
